using System;
using Microsoft.DirectX;

namespace PhysicsEngine
{
    public class MyMath
    {
        private MyMath()
        {
        }
        
        public static bool RelEqual(float a, float b, int decimals)
        {
            if (a == b) return true;
            if (a == 0) return (Math.Round(b, decimals) == 0);
            if (b == 0) return (Math.Round(a, decimals) == 0);
            return (Math.Round(a/b, decimals) == 1);
        }


        public static Matrix InvertMatrix3(Matrix m)
        {
            m.M44 = 1;
            m.Invert();
            m.M44 = 0;
            return m;
        }
    }

    public struct Matrix3
    {
        private Vector3 v1, v2, v3;
        
        public Matrix3(Vector3 v1, Vector3 v2, Vector3 v3)
        {
            this.v1 = v1;
            this.v2 = v2;
            this.v3 = v3;
        }
        
        // row major
        public float M11 { get { return v1.X; } set { v1.X = value; } }
        public float M12 { get { return v1.Y; } set { v1.Y = value; } }
        public float M13 { get { return v1.Z; } set { v1.Z = value; } }
        public float M21 { get { return v2.X; } set { v2.X = value; } }
        public float M22 { get { return v2.Y; } set { v2.Y = value; } }
        public float M23 { get { return v2.Z; } set { v2.Z = value; } }
        public float M31 { get { return v3.X; } set { v3.X = value; } }
        public float M32 { get { return v3.Y; } set { v3.Y = value; } }
        public float M33 { get { return v3.Z; } set { v3.Z = value; } }

        public static Matrix3 operator*(Matrix3 m, float f)
        {
            return new Matrix3(m.v1*f, m.v2*f, m.v3*f);
        }

        public static Matrix3 operator*(float f, Matrix3 m)
        {
            return new Matrix3(m.v1*f, m.v2*f, m.v3*f);
        }

        public static Vector3 operator*(Matrix3 m, Vector3 v)
        {
            return new Vector3(Vector3.Dot(m.v1,v), Vector3.Dot(m.v2,v), Vector3.Dot(m.v3,v));
        }

        public static Vector3 operator*(Vector3 v, Matrix3 m)
        {
            return new Vector3(v.X*m.M11+v.Y*m.M21+v.Z*m.M31, v.X*m.M12+v.Y*m.M22+v.Z*m.M32, v.X*m.M13+v.Y*m.M23+v.Z*m.M33);
        }

        public static Matrix3 operator+(Matrix3 m1, Matrix3 m2)
        {
            return new Matrix3(m1.v1+m2.v1, m1.v2+m2.v2, m1.v3+m2.v3);
        }

        public static Matrix3 operator-(Matrix3 m1, Matrix3 m2)
        {
            return new Matrix3(m1.v1-m2.v1, m1.v2-m2.v2, m1.v3-m2.v3);
        }

        public static Matrix3 operator*(Matrix3 m1, Matrix3 m2)
        {
            return new Matrix3(m1.v1*m2, m1.v2*m2, m1.v3*m2);
        }

        public Matrix3 Inverse()
        {
            return Matrix3.FromMatrix4(Matrix.Invert(Matrix3.ToMatrix4(this)));
        }

        public Matrix3 Transpose()
        {
            return new Matrix3(new Vector3(M11,M21,M31), new Vector3(M12,M22,M32), new Vector3(M13,M23,M33));
        }

        public static Matrix3 CrossProductMatrix(Vector3 v)
        {
            Matrix3 m = Matrix3.Zero;
            m.M12 = -v.Z;
            m.M13 = v.Y;
            m.M21 = v.Z;
            m.M23 = -v.X;
            m.M31 = -v.Y;
            m.M32 = v.X;
            
            return m;
        }

        public static Matrix ToMatrix4(Matrix3 m3)
        {
            Matrix m4 = Matrix.Identity;
            m4.M11 = m3.M11;
            m4.M21 = m3.M12;
            m4.M31 = m3.M13;
            m4.M12 = m3.M21;
            m4.M22 = m3.M22;
            m4.M32 = m3.M23;
            m4.M13 = m3.M31;
            m4.M23 = m3.M32;
            m4.M33 = m3.M33;
            return m4;
        }
                
        public static Matrix3 FromMatrix4(Matrix m4)
        {
            Matrix3 m3 = Matrix3.Zero;
            m3.M11 = m4.M11;
            m3.M21 = m4.M12;
            m3.M31 = m4.M13;
            m3.M12 = m4.M21;
            m3.M22 = m4.M22;
            m3.M32 = m4.M23;
            m3.M13 = m4.M31;
            m3.M23 = m4.M32;
            m3.M33 = m4.M33;
            return m3;
        }


        // static stuff

        public static readonly Matrix3 Identity;
        public static readonly Matrix3 Zero;

        static Matrix3()
        {
            Identity = new Matrix3(new Vector3(1,0,0), new Vector3(0,1,0), new Vector3(0,0,1));
            Zero = new Matrix3(new Vector3(0,0,0), new Vector3(0,0,0), new Vector3(0,0,0));
        }
    }
}
