using System;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace PhysicsEngine
{
	/// <summary>
	/// Summary description for SDMTest.
	/// </summary>
	public class SDMTest : GraphicsApplication
	{
		private const string meshFileName = 
            @"..\..\models\tetrahedron.x";
        
        private RigidBodyType bodyType;
        private RigidBody body;
        private Device device;
        private Bitmap bitmap;
        private VertexBuffer vertices;
        private Timer timer;

        private float scale;
        private int resolution;
        private float colorScale;
        
        private float lastTime;
        
        public SDMTest()
		{
            this.KeyDown += new KeyEventHandler(this.OnPrivateKeyDown);
            this.KeyUp += new KeyEventHandler(this.OnPrivateKeyUp);

            this.Text = "SDMTest";
            this.timer = new Timer();
            timer.Running = true;

            scale = 1;
            resolution = 256;
            colorScale = 256;
            lastTime = 0;
        }
        
        private void OnPrivateKeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Up:
                {
                    body.Vel = new Vector3(0, scale/2, 0);
                    break;
                }
                case Keys.Down:
                {
                    body.Vel = new Vector3(0, -scale/2, 0);
                    break;
                }
                case Keys.Left:
                {
                    body.Vel = new Vector3(-scale/2, 0, 0);
                    break;
                }
                case Keys.Right:
                {
                    body.Vel = new Vector3(scale/2, 0, 0);
                    break;
                }
                case Keys.PageUp:
                {
                    body.Vel = new Vector3(0, 0, scale/2);
                    break;
                }
                case Keys.PageDown:
                {
                    body.Vel = new Vector3(0, 0, -scale/2);
                    break;
                }
                case Keys.Home:
                {
                    scale *= 2;
                    break;
                }
                case Keys.End:
                {
                    scale *= .5f;
                    break;
                }
            }
        }

        private void OnPrivateKeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Up:
                case Keys.Down:
                case Keys.Left:
                case Keys.Right:
                case Keys.PageUp:
                case Keys.PageDown:
                {
                    body.Vel = Vector3.Empty;
                    break;
                }
            }
        }

        public override bool InitializeGraphics()
        {
            PresentParameters pres = new PresentParameters(); 
            pres.Windowed = true; 
            pres.SwapEffect = SwapEffect.Discard;

            device = new Device(0, DeviceType.Hardware, this, 
                CreateFlags.SoftwareVertexProcessing, pres); 

            bodyType = new RigidBodyType(null, meshFileName, device);
            body = new RigidBody(null, bodyType, new Vector3(), new Vector3(), new Quaternion(), new Vector3(), 0, 0, 1.0f);

            bitmap = null;

            vertices = new VertexBuffer( 
                typeof (CustomVertex.TransformedTextured), 
                4, device, 0, 
                CustomVertex.TransformedTextured.Format, 
                Pool.Default); 

            CustomVertex.TransformedTextured[] verts = 
                (CustomVertex.TransformedTextured[]) vertices.Lock(0, 0); 

            int i = 0; 
            verts[i++] = new CustomVertex.TransformedTextured( 
                0, resolution, 0.5F, 1,
                0, 0);

            verts[i++] = new CustomVertex.TransformedTextured( 
                0, 0, 0.5F, 1, 
                0, 1); 

            verts[i++] = new CustomVertex.TransformedTextured( 
                resolution, resolution, 0.5F, 1, 
                1, 0); 

            verts[i++] = new CustomVertex.TransformedTextured( 
                resolution, 0, 0.5F, 1, 
                1, 1); 

            vertices.Unlock(); 

            return true;
        }

        public override void DisposeGraphics()
        {
        }


        public override void Render()
        {
            device.Clear(ClearFlags.Target, Color.Bisque, 1.0f, 0);
    
            device.BeginScene();

            //body.UpdateVel(timer.GetElapsedTime());
            float curTime = timer.CurrentTime;
            body.UpdatePos(curTime - lastTime);
            lastTime = curTime;

            using (Texture texture = GenerateTexture())
            {
                device.SetTexture(0, texture); 

                device.VertexFormat = CustomVertex.TransformedTextured.Format; 
                device.SetStreamSource(0, vertices, 0); 
                device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);       

                device.EndScene();

                device.Present();
            }
        }

        private Texture GenerateTexture()
        {
            if (bitmap == null)
                bitmap = new Bitmap(resolution, resolution);

            for (int i = 0; i < resolution; i++)
            {
                for (int j = 0; j < resolution; j++)
                {
                    SDM.Value v = body.EvaluateSDM(body.PointWorldToBody(new Vector3((i - resolution/2)*scale/resolution, (j-resolution/2)*scale/resolution, 0)), true);
                    if (false) 
                    { // draw distance
                        if (v.d < 0)
                            bitmap.SetPixel(i, j, Color.FromArgb(255, Math.Min((int)(-v.d * colorScale), 255), Math.Min((int)(-v.d * colorScale), 255)));
                        else
                            bitmap.SetPixel(i, j, Color.FromArgb(Math.Min((int)(v.d * colorScale), 255), Math.Min((int)(v.d * colorScale), 255), 255));
                    } 
                    else 
                    { // draw normals
                        v.n = body.VectorBodyToWorld(v.n);
                        bitmap.SetPixel(i,j, Color.FromArgb((int)(Math.Abs(v.n.X)*255), (int)(Math.Abs(v.n.Y)*255), (int)(Math.Abs(v.n.Z)*255)));
                    }
                }
            }

            return Texture.FromBitmap(device, bitmap, Usage.Dynamic, Pool.Default);
        }
	}
}
